@@ -1,4 +1,3 @@
-import code
 import contextlib
 import errno
 import itertools
@@ -12,11 +11,20 @@ import tempfile
 import time
 import unicodedata
 from enum import Enum
-from types import TracebackType
-from typing import Dict, Any, List, Optional, Tuple, Union, cast, Type
-from typing_extensions import Literal
+from types import FrameType, TracebackType
+from typing import (
+    Any,
+    Iterable,
+    Dict,
+    List,
+    Literal,
+    Optional,
+    Sequence,
+    Tuple,
+    Type,
+    Union,
+)
 
-import blessings
 import greenlet
 from curtsies import (
     FSArray,
@@ -29,6 +37,7 @@ from curtsies import (
 )
 from curtsies.configfile_keynames import keymap as key_dispatch
 from curtsies.input import is_main_thread
+from curtsies.window import CursorAwareWindow
 from cwcwidth import wcswidth
 from pygments import format as pygformat
 from pygments.formatters import TerminalFormatter
@@ -46,7 +55,11 @@ from .interpreter import (
     Interp,
     code_finished_will_parse,
 )
-from .manual_readline import edit_keys
+from .manual_readline import (
+    edit_keys,
+    cursor_on_closing_char_pair,
+    AbstractEdits,
+)
 from .parse import parse as bpythonparse, func_for_letter, color_for_letter
 from .preprocess import preprocess
 from .. import __version__
@@ -58,6 +71,7 @@ from ..repl import (
     SourceNotFound,
 )
 from ..translations import _
+from ..line import CHARACTER_PAIR_MAP
 
 logger = logging.getLogger(__name__)
 
@@ -75,6 +89,12 @@ EDIT_SESSION_HEADER = """### current bpython session - make changes and save to
 MAX_EVENTS_POSSIBLY_NOT_PASTE = 20
 
 
+class SearchMode(Enum):
+    NO_SEARCH = 0
+    INCREMENTAL_SEARCH = 1
+    REVERSE_INCREMENTAL_SEARCH = 2
+
+
 class LineType(Enum):
     """Used when adding a tuple to all_logical_lines, to get input / output values
     having to actually type/know the strings"""
@@ -89,15 +109,20 @@ class FakeStdin:
     In user code, sys.stdin.read() asks the user for interactive input,
     so this class returns control to the UI to get that input."""
 
-    def __init__(self, coderunner, repl, configured_edit_keys=None):
+    def __init__(
+        self,
+        coderunner: CodeRunner,
+        repl: "BaseRepl",
+        configured_edit_keys: Optional[AbstractEdits] = None,
+    ):
         self.coderunner = coderunner
         self.repl = repl
         self.has_focus = False  # whether FakeStdin receives keypress events
         self.current_line = ""
         self.cursor_offset = 0
         self.old_num_lines = 0
-        self.readline_results = []
-        if configured_edit_keys:
+        self.readline_results: List[str] = []
+        if configured_edit_keys is not None:
             self.rl_char_sequences = configured_edit_keys
         else:
             self.rl_char_sequences = edit_keys
@@ -106,79 +131,94 @@ class FakeStdin:
         assert self.has_focus
 
         logger.debug("fake input processing event %r", e)
-        if isinstance(e, events.PasteEvent):
-            for ee in e.events:
-                if ee not in self.rl_char_sequences:
-                    self.add_input_character(ee)
+        if isinstance(e, events.Event):
+            if isinstance(e, events.PasteEvent):
+                for ee in e.events:
+                    if ee not in self.rl_char_sequences:
+                        self.add_input_character(ee)
+            elif isinstance(e, events.SigIntEvent):
+                self.coderunner.sigint_happened_in_main_context = True
+                self.has_focus = False
+                self.current_line = ""
+                self.cursor_offset = 0
+                self.repl.run_code_and_maybe_finish()
         elif e in self.rl_char_sequences:
             self.cursor_offset, self.current_line = self.rl_char_sequences[e](
                 self.cursor_offset, self.current_line
             )
-        elif isinstance(e, events.SigIntEvent):
-            self.coderunner.sigint_happened_in_main_context = True
-            self.has_focus = False
-            self.current_line = ""
-            self.cursor_offset = 0
-            self.repl.run_code_and_maybe_finish()
-        elif e in ("<ESC>",):
-            pass
-        elif e in ("<Ctrl-d>",):
-            if self.current_line == "":
+        elif e == "<Ctrl-d>":
+            if not len(self.current_line):
                 self.repl.send_to_stdin("\n")
                 self.has_focus = False
                 self.current_line = ""
                 self.cursor_offset = 0
                 self.repl.run_code_and_maybe_finish(for_code="")
-            else:
-                pass
         elif e in ("\n", "\r", "<Ctrl-j>", "<Ctrl-m>"):
-            line = self.current_line
-            self.repl.send_to_stdin(line + "\n")
+            line = f"{self.current_line}\n"
+            self.repl.send_to_stdin(line)
             self.has_focus = False
             self.current_line = ""
             self.cursor_offset = 0
-            self.repl.run_code_and_maybe_finish(for_code=line + "\n")
-        else:  # add normal character
+            self.repl.run_code_and_maybe_finish(for_code=line)
+        elif e != "<ESC>":  # add normal character
             self.add_input_character(e)
 
-        if self.current_line.endswith(("\n", "\r")):
-            pass
-        else:
+        if not self.current_line.endswith(("\n", "\r")):
             self.repl.send_to_stdin(self.current_line)
 
-    def add_input_character(self, e):
-        if e in ("<SPACE>",):
+    def add_input_character(self, e: str) -> None:
+        if e == "<SPACE>":
             e = " "
         if e.startswith("<") and e.endswith(">"):
             return
         assert len(e) == 1, "added multiple characters: %r" % e
         logger.debug("adding normal char %r to current line", e)
 
-        c = e
         self.current_line = (
             self.current_line[: self.cursor_offset]
-            + c
+            + e
             + self.current_line[self.cursor_offset :]
         )
         self.cursor_offset += 1
 
-    def readline(self):
+    def readline(self, size: int = -1) -> str:
+        if not isinstance(size, int):
+            raise TypeError(
+                f"'{type(size).__name__}' object cannot be interpreted as an integer"
+            )
+        elif size == 0:
+            return ""
         self.has_focus = True
         self.repl.send_to_stdin(self.current_line)
         value = self.coderunner.request_from_main_context()
+        assert isinstance(value, str)
         self.readline_results.append(value)
-        return value
+        return value if size <= -1 else value[:size]
 
-    def readlines(self, size=-1):
-        return list(iter(self.readline, ""))
+    def readlines(self, size: Optional[int] = -1) -> List[str]:
+        if size is None:
+            # the default readlines implementation also accepts None
+            size = -1
+        if not isinstance(size, int):
+            raise TypeError("argument should be integer or None, not 'str'")
+        if size <= 0:
+            # read as much as we can
+            return list(iter(self.readline, ""))
+
+        lines = []
+        while size > 0:
+            line = self.readline()
+            lines.append(line)
+            size -= len(line)
+        return lines
 
     def __iter__(self):
         return iter(self.readlines())
 
-    def isatty(self):
+    def isatty(self) -> bool:
         return True
 
-    def flush(self):
+    def flush(self) -> None:
         """Flush the internal buffer. This is a no-op. Flushing stdin
         doesn't make any sense anyway."""
 
@@ -187,7 +227,7 @@ class FakeStdin:
         # others, so here's a hack to keep them happy
         raise OSError(errno.EBADF, "sys.stdin is read-only")
 
-    def close(self):
+    def close(self) -> None:
         # hack to make closing stdin a nop
         # This is useful for multiprocessing.Process, which does work
         # for the most part, although output from other processes is
@@ -195,8 +235,9 @@ class FakeStdin:
         pass
 
     @property
-    def encoding(self):
-        return sys.__stdin__.encoding
+    def encoding(self) -> str:
+        # `encoding` is new in py39
+        return sys.__stdin__.encoding  # type: ignore
 
     # TODO write a read() method?
 
@@ -205,7 +246,7 @@ class ReevaluateFakeStdin:
     """Stdin mock used during reevaluation (undo) so raw_inputs don't have to
     be reentered"""
 
-    def __init__(self, fakestdin, repl):
+    def __init__(self, fakestdin: FakeStdin, repl: "BaseRepl"):
         self.fakestdin = fakestdin
         self.repl = repl
         self.readline_results = fakestdin.readline_results[:]
@@ -229,54 +270,39 @@ class ImportLoader:
     def __getattr__(self, name):
         if name == "create_module" and hasattr(self.loader, name):
             return self._create_module
-        if name == "load_module" and hasattr(self.loader, name):
-            return self._load_module
         return getattr(self.loader, name)
 
     def _create_module(self, spec):
-        spec = self.loader.create_module(spec)
+        module_object = self.loader.create_module(spec)
         if (
             getattr(spec, "origin", None) is not None
             and spec.origin != "builtin"
         ):
             self.watcher.track_module(spec.origin)
-        return spec
-
-    def _load_module(self, name):
-        module = self.loader.load_module(name)
-        if hasattr(module, "__file__"):
-            self.watcher.track_module(module.__file__)
-        return module
+        return module_object
 
 
 class ImportFinder:
-    """Wrapper for finders in sys.meta_path to replace wrap all loaders with ImportLoader."""
+    """Wrapper for finders in sys.meta_path to wrap all loaders with ImportLoader."""
 
-    def __init__(self, finder, watcher):
+    def __init__(self, watcher, finder):
         self.watcher = watcher
         self.finder = finder
 
     def __getattr__(self, name):
         if name == "find_spec" and hasattr(self.finder, name):
             return self._find_spec
-        if name == "find_module" and hasattr(self.finder, name):
-            return self._find_module
         return getattr(self.finder, name)
 
     def _find_spec(self, fullname, path, target=None):
         # Attempt to find the spec
         spec = self.finder.find_spec(fullname, path, target)
         if spec is not None:
-            if getattr(spec, "__loader__", None) is not None:
+            if getattr(spec, "loader", None) is not None:
                 # Patch the loader to enable reloading
-                spec.__loader__ = ImportLoader(self.watcher, spec.__loader__)
+                spec.loader = ImportLoader(self.watcher, spec.loader)
         return spec
 
-    def _find_module(self, fullname, path=None):
-        loader = self.finder.find_module(fullname, path)
-        if loader is not None:
-            return ImportLoader(self.watcher, loader)
-
 
 def _process_ps(ps, default_ps: str):
     """Replace ps1/ps2 with the default if the user specified value contains control characters."""
@@ -311,10 +337,11 @@ class BaseRepl(Repl):
     def __init__(
         self,
         config: Config,
-        locals_: Dict[str, Any] = None,
-        banner: str = None,
-        interp: code.InteractiveInterpreter = None,
-        orig_tcattrs: List[Any] = None,
+        window: CursorAwareWindow,
+        locals_: Optional[Dict[str, Any]] = None,
+        banner: Optional[str] = None,
+        interp: Optional[Interp] = None,
+        orig_tcattrs: Optional[List[Any]] = None,
     ):
         """
         locals_ is a mapping of locals to pass into the interpreter
@@ -325,6 +352,7 @@ class BaseRepl(Repl):
         """
 
         logger.debug("starting init")
+        self.window = window
 
         # If creating a new interpreter on undo would be unsafe because initial
         # state was passed in
@@ -356,6 +384,8 @@ class BaseRepl(Repl):
         )
         self.edit_keys = edit_keys.mapping_with_config(config, key_dispatch)
         logger.debug("starting parent init")
+        # interp is a subclass of repl.Interpreter, so it definitely,
+        # implements the methods of Interpreter!
         super().__init__(interp, config)
 
         self.formatter = BPythonFormatter(config.color_scheme)
@@ -369,7 +399,7 @@ class BaseRepl(Repl):
         self._current_line = ""
 
         # current line of output - stdout and stdin go here
-        self.current_stdouterr_line = ""  # Union[str, FmtStr]
+        self.current_stdouterr_line: Union[str, FmtStr] = ""
 
         # this is every line that's been displayed (input and output)
         # as with formatting applied. Logical lines that exceeded the terminal width
@@ -441,17 +471,19 @@ class BaseRepl(Repl):
         # whether auto reloading active
         self.watching_files = config.default_autoreload
 
-        # 'reverse_incremental_search', 'incremental_search' or None
-        self.incr_search_mode = None
-
+        self.incr_search_mode = SearchMode.NO_SEARCH
         self.incr_search_target = ""
 
         self.original_modules = set(sys.modules.keys())
 
         # as long as the first event received is a window resize event,
         # this works fine...
-        self.width: int = cast(int, None)
-        self.height: int = cast(int, None)
+        try:
+            self.width, self.height = os.get_terminal_size()
+        except OSError:
+            # this case will trigger during unit tests when stdout is redirected
+            self.width = -1
+            self.height = -1
 
         self.status_bar.message(banner)
 
@@ -498,7 +530,7 @@ class BaseRepl(Repl):
         RefreshRequestEvent."""
         raise NotImplementedError
 
-    def _request_reload(self, files_modified=("?",)):
+    def _request_reload(self, files_modified: Sequence[str]) -> None:
         """Like request_refresh, but for reload requests events."""
         raise NotImplementedError
 
@@ -528,15 +560,15 @@ class BaseRepl(Repl):
         else:
             self._request_refresh()
 
-    def request_reload(self, files_modified=()):
+    def request_reload(self, files_modified: Sequence[str] = ()) -> None:
         """Request that a ReloadEvent be passed next into process_event"""
         if self.watching_files:
-            self._request_reload(files_modified=files_modified)
+            self._request_reload(files_modified)
 
-    def schedule_refresh(self, when="now"):
+    def schedule_refresh(self, when: float = 0) -> None:
         """Schedule a ScheduledRefreshRequestEvent for when.
 
-        Such a event should interrupt if blockied waiting for keyboard input"""
+        Such a event should interrupt if blocked waiting for keyboard input"""
         if self.reevaluating or self.paste_mode:
             self.fake_refresh_requested = True
         else:
@@ -561,14 +593,7 @@ class BaseRepl(Repl):
         if self.watcher:
             meta_path = []
             for finder in sys.meta_path:
-                # All elements get wrapped in ImportFinder instances execepted for instances of
-                # _SixMetaPathImporter (from six). When importing six, it will check if the importer
-                # is already part of sys.meta_path and will remove instances. We do not want to
-                # break this feature (see also #874).
-                if type(finder).__name__ == "_SixMetaPathImporter":
-                    meta_path.append(finder)
-                else:
-                    meta_path.append(ImportFinder(finder, self.watcher))
+                meta_path.append(ImportFinder(self.watcher, finder))
             sys.meta_path = meta_path
 
         sitefix.monkeypatch_quit()
@@ -592,7 +617,7 @@ class BaseRepl(Repl):
         sys.meta_path = self.orig_meta_path
         return False
 
-    def sigwinch_handler(self, signum, frame):
+    def sigwinch_handler(self, signum: int, frame: Optional[FrameType]) -> None:
         old_rows, old_columns = self.height, self.width
         self.height, self.width = self.get_term_hw()
         cursor_dy = self.get_cursor_vertical_diff()
@@ -608,9 +633,9 @@ class BaseRepl(Repl):
             self.scroll_offset,
         )
 
-    def sigtstp_handler(self, signum, frame):
+    def sigtstp_handler(self, signum: int, frame: Optional[FrameType]) -> None:
         self.scroll_offset = len(self.lines_for_display)
-        self.__exit__()
+        self.__exit__(None, None, None)
         self.on_suspend()
         os.kill(os.getpid(), signal.SIGTSTP)
         self.after_suspend()
@@ -636,8 +661,7 @@ class BaseRepl(Repl):
             self.process_key_event(e)
             return None
 
-    def process_control_event(self, e) -> Optional[bool]:
-
+    def process_control_event(self, e: events.Event) -> Optional[bool]:
         if isinstance(e, bpythonevents.ScheduledRefreshRequestEvent):
             # This is a scheduled refresh - it's really just a refresh (so nop)
             pass
@@ -682,9 +706,9 @@ class BaseRepl(Repl):
         elif isinstance(e, bpythonevents.RunStartupFileEvent):
             try:
                 self.startup()
-            except OSError as e:
+            except OSError as err:
                 self.status_bar.message(
-                    _("Executing PYTHONSTARTUP failed: %s") % (e,)
+                    _("Executing PYTHONSTARTUP failed: %s") % (err,)
                 )
 
         elif isinstance(e, bpythonevents.UndoEvent):
@@ -735,11 +759,11 @@ class BaseRepl(Repl):
             self.up_one_line()
         elif e in ("<DOWN>",) + key_dispatch[self.config.down_one_line_key]:
             self.down_one_line()
-        elif e in ("<Ctrl-d>",):
+        elif e == "<Ctrl-d>":
             self.on_control_d()
-        elif e in ("<Ctrl-o>",):
+        elif e == "<Ctrl-o>":
             self.operate_and_get_next()
-        elif e in ("<Esc+.>",):
+        elif e == "<Esc+.>":
             self.get_last_word()
         elif e in key_dispatch[self.config.reverse_incremental_search_key]:
             self.incremental_search(reverse=True)
@@ -747,7 +771,7 @@ class BaseRepl(Repl):
             self.incremental_search()
         elif (
             e in (("<BACKSPACE>",) + key_dispatch[self.config.backspace_key])
-            and self.incr_search_mode
+            and self.incr_search_mode != SearchMode.NO_SEARCH
         ):
             self.add_to_incremental_search(self, backspace=True)
         elif e in self.edit_keys.cut_buffer_edits:
@@ -775,9 +799,9 @@ class BaseRepl(Repl):
             raise SystemExit()
         elif e in ("\n", "\r", "<PADENTER>", "<Ctrl-j>", "<Ctrl-m>"):
             self.on_enter()
-        elif e in ("<TAB>",):  # tab
+        elif e == "<TAB>":  # tab
             self.on_tab()
-        elif e in ("<Shift-TAB>",):
+        elif e == "<Shift-TAB>":
             self.on_tab(back=True)
         elif e in key_dispatch[self.config.undo_key]:  # ctrl-r for undo
             self.prompt_undo()
@@ -796,15 +820,80 @@ class BaseRepl(Repl):
         # TODO add PAD keys hack as in bpython.cli
         elif e in key_dispatch[self.config.edit_current_block_key]:
             self.send_current_block_to_external_editor()
-        elif e in ("<ESC>",):
-            self.incr_search_mode = None
-        elif e in ("<SPACE>",):
+        elif e == "<ESC>":
+            self.incr_search_mode = SearchMode.NO_SEARCH
+        elif e == "<SPACE>":
             self.add_normal_character(" ")
+        elif e in CHARACTER_PAIR_MAP.keys():
+            if e in ["'", '"']:
+                if self.is_closing_quote(e):
+                    self.insert_char_pair_end(e)
+                else:
+                    self.insert_char_pair_start(e)
+            else:
+                self.insert_char_pair_start(e)
+        elif e in CHARACTER_PAIR_MAP.values():
+            self.insert_char_pair_end(e)
         else:
             self.add_normal_character(e)
 
-    def get_last_word(self):
+    def is_closing_quote(self, e: str) -> bool:
+        char_count = self._current_line.count(e)
+        return (
+            char_count % 2 == 0
+            and cursor_on_closing_char_pair(
+                self._cursor_offset, self._current_line, e
+            )[0]
+        )
 
+    def insert_char_pair_start(self, e):
+        """Accepts character which is a part of CHARACTER_PAIR_MAP
+        like brackets and quotes, and appends it to the line with
+        an appropriate character pair ending. Closing character can only be inserted
+        when the next character is either a closing character or a space
+
+        e.x. if you type "(" (lparen) , this will insert "()"
+        into the line
+        """
+        self.add_normal_character(e)
+        if self.config.brackets_completion:
+            start_of_line = len(self._current_line) == 1
+            end_of_line = len(self._current_line) == self._cursor_offset
+            can_lookup_next = len(self._current_line) > self._cursor_offset
+            next_char = (
+                None
+                if not can_lookup_next
+                else self._current_line[self._cursor_offset]
+            )
+            if (
+                start_of_line
+                or end_of_line
+                or (next_char is not None and next_char in "})] ")
+            ):
+                self.add_normal_character(
+                    CHARACTER_PAIR_MAP[e], narrow_search=False
+                )
+                self._cursor_offset -= 1
+
+    def insert_char_pair_end(self, e):
+        """Accepts character which is a part of CHARACTER_PAIR_MAP
+        like brackets and quotes, and checks whether it should be
+        inserted to the line or overwritten
+
+        e.x. if you type ")" (rparen) , and your cursor is directly
+        above another ")" (rparen) in the cmd, this will just skip
+        it and move the cursor.
+        If there is no same character underneath the cursor, the
+        character will be printed/appended to the line
+        """
+        if self.config.brackets_completion:
+            if self.cursor_offset < len(self._current_line):
+                if self._current_line[self.cursor_offset] == e:
+                    self.cursor_offset += 1
+                    return
+        self.add_normal_character(e)
+
+    def get_last_word(self):
         previous_word = _last_word(self.rl_history.entry)
         word = _last_word(self.rl_history.back())
         line = self.current_line
@@ -818,7 +907,7 @@ class BaseRepl(Repl):
         )
 
     def incremental_search(self, reverse=False, include_current=False):
-        if self.incr_search_mode is None:
+        if self.incr_search_mode == SearchMode.NO_SEARCH:
             self.rl_history.enter(self.current_line)
             self.incr_search_target = ""
         else:
@@ -847,9 +936,9 @@ class BaseRepl(Repl):
                     clear_special_mode=False,
                 )
         if reverse:
-            self.incr_search_mode = "reverse_incremental_search"
+            self.incr_search_mode = SearchMode.REVERSE_INCREMENTAL_SEARCH
         else:
-            self.incr_search_mode = "incremental_search"
+            self.incr_search_mode = SearchMode.INCREMENTAL_SEARCH
 
     def readline_kill(self, e):
         func = self.edit_keys[e]
@@ -894,7 +983,7 @@ class BaseRepl(Repl):
             """returns true if all characters before cursor are whitespace"""
             return not self.current_line[: self.cursor_offset].strip()
 
-        logger.debug("self.matches_iter.matches:%r", self.matches_iter.matches)
+        logger.debug("self.matches_iter.matches: %r", self.matches_iter.matches)
         if only_whitespace_left_of_cursor():
             front_ws = len(self.current_line[: self.cursor_offset]) - len(
                 self.current_line[: self.cursor_offset].lstrip()
@@ -903,7 +992,15 @@ class BaseRepl(Repl):
             for unused in range(to_add):
                 self.add_normal_character(" ")
             return
-
+        # if cursor on closing character from pair,
+        # moves cursor behind it on tab
+        # ? should we leave it here as default?
+        if self.config.brackets_completion:
+            on_closing_char, _ = cursor_on_closing_char_pair(
+                self._cursor_offset, self._current_line
+            )
+            if on_closing_char:
+                self._cursor_offset += 1
         # run complete() if we don't already have matches
         if len(self.matches_iter.matches) == 0:
             self.list_win_visible = self.complete(tab=True)
@@ -915,7 +1012,6 @@ class BaseRepl(Repl):
             # using _current_line so we don't trigger a completion reset
             if not self.matches_iter.matches:
                 self.list_win_visible = self.complete()
-
         elif self.matches_iter.matches:
             self.current_match = (
                 back and self.matches_iter.previous() or next(self.matches_iter)
@@ -924,6 +1020,24 @@ class BaseRepl(Repl):
             self._cursor_offset, self._current_line = cursor_and_line
             # using _current_line so we don't trigger a completion reset
             self.list_win_visible = True
+        if self.config.brackets_completion:
+            # appends closing char pair if completion is a callable
+            if self.is_completion_callable(self._current_line):
+                self._current_line = self.append_closing_character(
+                    self._current_line
+                )
+
+    def is_completion_callable(self, completion):
+        """Checks whether given completion is callable (e.x. function)"""
+        completion_end = completion[-1]
+        return completion_end in CHARACTER_PAIR_MAP
+
+    def append_closing_character(self, completion):
+        """Appends closing character/bracket to the completion"""
+        completion_end = completion[-1]
+        if completion_end in CHARACTER_PAIR_MAP:
+            completion = f"{completion}{CHARACTER_PAIR_MAP[completion_end]}"
+        return completion
 
     def on_control_d(self):
         if self.current_line == "":
@@ -972,7 +1086,7 @@ class BaseRepl(Repl):
         )
         self._set_cursor_offset(len(self.current_line), reset_rl_history=False)
 
-    def process_simple_keypress(self, e):
+    def process_simple_keypress(self, e: str):
         # '\n' needed for pastes
         if e in ("<Ctrl-j>", "<Ctrl-m>", "<PADENTER>", "\n", "\r"):
             self.on_enter()
@@ -981,7 +1095,7 @@ class BaseRepl(Repl):
                 self.process_event(bpythonevents.RefreshRequestEvent())
         elif isinstance(e, events.Event):
             pass  # ignore events
-        elif e in ("<SPACE>",):
+        elif e == "<SPACE>":
             self.add_normal_character(" ")
         else:
             self.add_normal_character(e)
@@ -1071,10 +1185,10 @@ class BaseRepl(Repl):
             )
 
     # Handler Helpers
-    def add_normal_character(self, char):
+    def add_normal_character(self, char, narrow_search=True):
         if len(char) > 1 or is_nop(char):
             return
-        if self.incr_search_mode:
+        if self.incr_search_mode != SearchMode.NO_SEARCH:
             self.add_to_incremental_search(char)
         else:
             self._set_current_line(
@@ -1087,27 +1201,33 @@ class BaseRepl(Repl):
                 reset_rl_history=False,
                 clear_special_mode=False,
             )
-            self.cursor_offset += 1
+            if narrow_search:
+                self.cursor_offset += 1
+            else:
+                self._cursor_offset += 1
         if self.config.cli_trim_prompts and self.current_line.startswith(
             self.ps1
         ):
             self.current_line = self.current_line[4:]
-            self.cursor_offset = max(0, self.cursor_offset - 4)
+            if narrow_search:
+                self.cursor_offset = max(0, self.cursor_offset - 4)
+            else:
+                self._cursor_offset += max(0, self.cursor_offset - 4)
 
     def add_to_incremental_search(self, char=None, backspace=False):
         """Modify the current search term while in incremental search.
 
         The only operations allowed in incremental search mode are
         adding characters and backspacing."""
-        if char is None and not backspace:
-            raise ValueError("must provide a char or set backspace to True")
         if backspace:
             self.incr_search_target = self.incr_search_target[:-1]
-        else:
+        elif char is not None:
             self.incr_search_target += char
-        if self.incr_search_mode == "reverse_incremental_search":
+        else:
+            raise ValueError("must provide a char or set backspace to True")
+        if self.incr_search_mode == SearchMode.REVERSE_INCREMENTAL_SEARCH:
             self.incremental_search(reverse=True, include_current=True)
-        elif self.incr_search_mode == "incremental_search":
+        elif self.incr_search_mode == SearchMode.INCREMENTAL_SEARCH:
             self.incremental_search(include_current=True)
         else:
             raise ValueError("add_to_incremental_search not in a special mode")
@@ -1132,7 +1252,9 @@ class BaseRepl(Repl):
         elif (
             line
             and ":" not in line
-            and line.strip().startswith(("return", "pass", "raise", "yield"))
+            and line.strip().startswith(
+                ("return", "pass", "...", "raise", "yield", "break", "continue")
+            )
         ):
             indent = max(0, indent - self.config.tab_length)
         logger.debug("indent we found was %s", indent)
@@ -1244,9 +1366,8 @@ class BaseRepl(Repl):
     def clear_current_block(self, remove_from_history=True):
         self.display_buffer = []
         if remove_from_history:
-            for unused in self.buffer:
-                self.history.pop()
-                self.all_logical_lines.pop()
+            del self.history[-len(self.buffer) :]
+            del self.all_logical_lines[-len(self.buffer) :]
         self.buffer = []
         self.cursor_offset = 0
         self.saved_indent = 0
@@ -1315,7 +1436,7 @@ class BaseRepl(Repl):
             fs = bpythonparse(
                 pygformat(self.tokenize(self.current_line), self.formatter)
             )
-            if self.incr_search_mode:
+            if self.incr_search_mode != SearchMode.NO_SEARCH:
                 if self.incr_search_target in self.current_line:
                     fs = fmtfuncs.on_magenta(self.incr_search_target).join(
                         fs.split(self.incr_search_target)
@@ -1363,12 +1484,12 @@ class BaseRepl(Repl):
         """colored line with prompt"""
         prompt = func_for_letter(self.config.color_scheme["prompt"])
         more = func_for_letter(self.config.color_scheme["prompt_more"])
-        if self.incr_search_mode == "reverse_incremental_search":
+        if self.incr_search_mode == SearchMode.REVERSE_INCREMENTAL_SEARCH:
             return (
                 prompt(f"(reverse-i-search)`{self.incr_search_target}': ")
                 + self.current_line_formatted
             )
-        elif self.incr_search_mode == "incremental_search":
+        elif self.incr_search_mode == SearchMode.INCREMENTAL_SEARCH:
             return prompt(f"(i-search)`%s': ") + self.current_line_formatted
         return (
             prompt(self.ps1) if self.done else more(self.ps2)
@@ -1424,15 +1545,15 @@ class BaseRepl(Repl):
 
         Should return zero unless there are fullwidth characters."""
         full_line = self.current_cursor_line_without_suggestion
-        line_with_padding = "".join(
-            line.s
+        line_with_padding_len = sum(
+            len(line.s)
             for line in paint.display_linize(
                 self.current_cursor_line_without_suggestion.s, self.width
             )
         )
 
         # the difference in length here is how much padding there is
-        return len(line_with_padding) - len(full_line)
+        return line_with_padding_len - len(full_line)
 
     def paint(
         self,
@@ -1681,9 +1802,11 @@ class BaseRepl(Repl):
                 self.current_match,
                 self.docstring,
                 self.config,
-                self.matches_iter.completer.format
-                if self.matches_iter.completer
-                else None,
+                (
+                    self.matches_iter.completer.format
+                    if self.matches_iter.completer
+                    else None
+                ),
             )
 
             if (
@@ -1757,18 +1880,13 @@ class BaseRepl(Repl):
   lines scrolled down: {self.scroll_offset}
 >"""
 
-    @property
-    def current_line(self):
+    def _get_current_line(self) -> str:
         """The current line"""
         return self._current_line
 
-    @current_line.setter
-    def current_line(self, value):
-        self._set_current_line(value)
-
     def _set_current_line(
         self,
-        line,
+        line: str,
         update_completion=True,
         reset_rl_history=True,
         clear_special_mode=True,
@@ -1786,18 +1904,13 @@ class BaseRepl(Repl):
             self.special_mode = None
         self.unhighlight_paren()
 
-    @property
-    def cursor_offset(self):
+    def _get_cursor_offset(self) -> int:
         """The current cursor offset from the front of the "line"."""
         return self._cursor_offset
 
-    @cursor_offset.setter
-    def cursor_offset(self, value):
-        self._set_cursor_offset(value)
-
     def _set_cursor_offset(
         self,
-        offset,
+        offset: int,
         update_completion=True,
         reset_rl_history=False,
         clear_special_mode=True,
@@ -1811,7 +1924,7 @@ class BaseRepl(Repl):
         if reset_rl_history:
             self.rl_history.reset()
         if clear_special_mode:
-            self.incr_search_mode = None
+            self.incr_search_mode = SearchMode.NO_SEARCH
         self._cursor_offset = offset
         if update_completion:
             self.update_completion()
@@ -1873,7 +1986,7 @@ class BaseRepl(Repl):
 
         greenlet.greenlet(prompt_for_undo).switch()
 
-    def redo(self):
+    def redo(self) -> None:
         if self.redo_stack:
             temp = self.redo_stack.pop()
             self.history.append(temp)
@@ -1954,7 +2067,7 @@ class BaseRepl(Repl):
 
         del self.coderunner.interp.locals["_repl"]
 
-    def getstdout(self):
+    def getstdout(self) -> str:
         """
         Returns a string of the current bpython session, wrapped, WITH prompts.
         """
@@ -1971,7 +2084,7 @@ class BaseRepl(Repl):
         try:
             signal.signal(signal.SIGWINCH, self.orig_sigwinch_handler)
             with Termmode(self.orig_stdin, self.orig_tcattrs):
-                terminal = blessings.Terminal(stream=sys.__stdout__)
+                terminal = self.window.t
                 with terminal.fullscreen():
                     sys.__stdout__.write(terminal.save)
                     sys.__stdout__.write(terminal.move(0, 0))
@@ -1988,17 +2101,17 @@ class BaseRepl(Repl):
         finally:
             signal.signal(signal.SIGWINCH, prev_sigwinch_handler)
 
-    def pager(self, text):
-        """Runs an external pager on text
+    def pager(self, text: str, title: str = "") -> None:
+        """Runs an external pager on text"""
 
-        text must be a str"""
+        # TODO: make less handle title
         command = get_pager_command()
         with tempfile.NamedTemporaryFile() as tmp:
             tmp.write(text.encode(getpreferredencoding()))
             tmp.flush()
             self.focus_on_subprocess(command + [tmp.name])
 
-    def show_source(self):
+    def show_source(self) -> None:
         try:
             source = self.get_source_of_current_name()
         except SourceNotFound as e:
@@ -2010,10 +2123,10 @@ class BaseRepl(Repl):
                 )
             self.pager(source)
 
-    def help_text(self):
+    def help_text(self) -> str:
         return self.version_help_text() + "\n" + self.key_help_text()
 
-    def version_help_text(self):
+    def version_help_text(self) -> str:
         help_message = _(
             """
 Thanks for using bpython!
@@ -2040,7 +2153,7 @@ Press {config.edit_config_key} to edit this config file.
 
         return f"bpython-curtsies version {__version__} using curtsies version {curtsies_version}\n{help_message}"
 
-    def key_help_text(self):
+    def key_help_text(self) -> str:
         NOT_IMPLEMENTED = (
             "suspend",
             "cut to buffer",
@@ -2067,7 +2180,7 @@ Press {config.edit_config_key} to edit this config file.
 
         max_func = max(len(func) for func, key in pairs)
         return "\n".join(
-            "{} : {}".format(func.rjust(max_func), key) for func, key in pairs
+            f"{func.rjust(max_func)} : {key}" for func, key in pairs
         )
 
     def get_session_formatted_for_file(self) -> str:
@@ -2090,15 +2203,15 @@ Press {config.edit_config_key} to edit this config file.
         return _process_ps(super().ps2, "... ")
 
 
-def is_nop(char):
-    return unicodedata.category(str(char)) == "Cc"
+def is_nop(char: str) -> bool:
+    return unicodedata.category(char) == "Cc"
 
 
-def tabs_to_spaces(line):
+def tabs_to_spaces(line: str) -> str:
     return line.replace("\t", "    ")
 
 
-def _last_word(line):
+def _last_word(line: str) -> str:
     split_line = line.split()
     return split_line.pop() if split_line else ""
 
@@ -2122,29 +2235,29 @@ def compress_paste_event(paste_event):
         return None
 
 
-def just_simple_events(event_list):
+def just_simple_events(
+    event_list: Iterable[Union[str, events.Event]]
+) -> List[str]:
     simple_events = []
     for e in event_list:
+        if isinstance(e, events.Event):
+            continue  # ignore events
         # '\n' necessary for pastes
-        if e in ("<Ctrl-j>", "<Ctrl-m>", "<PADENTER>", "\n", "\r"):
+        elif e in ("<Ctrl-j>", "<Ctrl-m>", "<PADENTER>", "\n", "\r"):
             simple_events.append("\n")
-        elif isinstance(e, events.Event):
-            pass  # ignore events
-        elif e in ("<SPACE>",):
+        elif e == "<SPACE>":
             simple_events.append(" ")
         elif len(e) > 1:
-            pass  # get rid of <Ctrl-a> etc.
+            continue  # get rid of <Ctrl-a> etc.
         else:
             simple_events.append(e)
     return simple_events
 
 
-def is_simple_event(e):
+def is_simple_event(e: Union[str, events.Event]) -> bool:
     if isinstance(e, events.Event):
         return False
-    if e in ("<Ctrl-j>", "<Ctrl-m>", "<PADENTER>", "\n", "\r", "<SPACE>"):
-        return True
-    if len(e) > 1:
-        return False
-    else:
-        return True
+    return (
+        e in ("<Ctrl-j>", "<Ctrl-m>", "<PADENTER>", "\n", "\r", "<SPACE>")
+        or len(e) <= 1
+    )
