@@ -5,7 +5,7 @@ from curtsies import fsarray, fmtstr, FSArray
 from curtsies.formatstring import linesplit
 from curtsies.fmtfuncs import bold
 
-from bpython.curtsiesfrontend.parse import func_for_letter
+from .parse import func_for_letter
 
 logger = logging.getLogger(__name__)
 
@@ -74,9 +74,11 @@ def matches_lines(rows, columns, matches, current, config, match_format):
 
     result = [
         fmtstr(" ").join(
-            color(m.ljust(max_match_width))
-            if m != current
-            else highlight_color(m.ljust(max_match_width))
+            (
+                color(m.ljust(max_match_width))
+                if m != current
+                else highlight_color(m.ljust(max_match_width))
+            )
             for m in matches[i : i + words_wide]
         )
         for i in range(0, len(matches), words_wide)
