@@ -21,6 +21,7 @@
 # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 # THE SOFTWARE.
 
+import abc
 import code
 import inspect
 import os
@@ -34,14 +35,27 @@ import textwrap
 import time
 import traceback
 from abc import abstractmethod
+from dataclasses import dataclass
 from itertools import takewhile
 from pathlib import Path
 from types import ModuleType, TracebackType
-from typing import cast, Tuple, Any, Optional, Type
-from typing_extensions import Literal
+from typing import (
+    Any,
+    Callable,
+    Dict,
+    Iterable,
+    List,
+    Literal,
+    Optional,
+    TYPE_CHECKING,
+    Tuple,
+    Type,
+    Union,
+    cast,
+)
 
 from pygments.lexers import Python3Lexer
-from pygments.token import Token
+from pygments.token import Token, _TokenType
 
 have_pyperclip = True
 try:
@@ -50,7 +64,7 @@ except ImportError:
     have_pyperclip = False
 
 from . import autocomplete, inspection, simpleeval
-from .config import getpreferredencoding
+from .config import getpreferredencoding, Config
 from .formatter import Parenthesis
 from .history import History
 from .lazyre import LazyReCompile
@@ -63,12 +77,11 @@ from .importcompletion import ModuleGatherer
 class RuntimeTimer:
     """Calculate running time"""
 
-    def __init__(self):
+    def __init__(self) -> None:
         self.reset_timer()
-        self.time = time.monotonic if hasattr(time, "monotonic") else time.time
 
-    def __enter__(self):
-        self.start = self.time()
+    def __enter__(self) -> None:
+        self.start = time.monotonic()
 
     def __exit__(
         self,
@@ -76,15 +89,15 @@ class RuntimeTimer:
         exc_val: Optional[BaseException],
         exc_tb: Optional[TracebackType],
     ) -> Literal[False]:
-        self.last_command = self.time() - self.start
+        self.last_command = time.monotonic() - self.start
         self.running_time += self.last_command
         return False
 
-    def reset_timer(self):
+    def reset_timer(self) -> None:
         self.running_time = 0.0
         self.last_command = 0.0
 
-    def estimate(self):
+    def estimate(self) -> float:
         return self.running_time - self.last_command
 
 
@@ -93,7 +106,10 @@ class Interpreter(code.InteractiveInterpreter):
 
     bpython_input_re = LazyReCompile(r"<bpython-input-\d+>")
 
-    def __init__(self, locals=None, encoding=None):
+    def __init__(
+        self,
+        locals: Optional[Dict[str, Any]] = None,
+    ) -> None:
         """Constructor.
 
         The optional 'locals' argument specifies the dictionary in which code
@@ -107,15 +123,9 @@ class Interpreter(code.InteractiveInterpreter):
         callback can be added to the Interpreter instance afterwards - more
         specifically, this is so that autoindentation does not occur after a
         traceback.
-
-        encoding is only used in Python 2, where it may be necessary to add an
-        encoding comment to a source bytestring before running it.
-        encoding must be a bytestring in Python 2 because it will be templated
-        into a bytestring source as part of an encoding comment.
         """
 
-        self.encoding = encoding or getpreferredencoding()
-        self.syntaxerror_callback = None
+        self.syntaxerror_callback: Optional[Callable] = None
 
         if locals is None:
             # instead of messing with sys.modules, we should modify sys.modules
@@ -126,41 +136,25 @@ class Interpreter(code.InteractiveInterpreter):
         super().__init__(locals)
         self.timer = RuntimeTimer()
 
-    def reset_running_time(self):
-        self.running_time = 0
-
-    def runsource(self, source, filename=None, symbol="single", encode="auto"):
+    def runsource(
+        self,
+        source: str,
+        filename: Optional[str] = None,
+        symbol: str = "single",
+    ) -> bool:
         """Execute Python code.
 
         source, filename and symbol are passed on to
-        code.InteractiveInterpreter.runsource. If encode is True,
-        an encoding comment will be added to the source.
-        On Python 3.X, encode will be ignored.
-
-        encode should only be used for interactive interpreter input,
-        files should always already have an encoding comment or be ASCII.
-        By default an encoding line will be added if no filename is given.
-
-        source must be a string
-
-        Because adding an encoding comment to a unicode string in Python 2
-        would cause a syntax error to be thrown which would reference code
-        the user did not write, setting encoding to True when source is a
-        unicode string in Python 2 will throw a ValueError."""
-        if encode and filename is not None:
-            # files have encoding comments or implicit encoding of ASCII
-            if encode != "auto":
-                raise ValueError("shouldn't add encoding line to file contents")
-            encode = False
+        code.InteractiveInterpreter.runsource."""
 
         if filename is None:
             filename = filename_for_console_input(source)
         with self.timer:
-            return code.InteractiveInterpreter.runsource(
-                self, source, filename, symbol
-            )
+            return super().runsource(source, filename, symbol)
 
-    def showsyntaxerror(self, filename=None):
+    def showsyntaxerror(
+        self, filename: Optional[str] = None, **kwargs
+    ) -> None:
         """Override the regular handler, the code's copied and pasted from
         code.py, as per showtraceback, but with the syntaxerror callback called
         and the text in a pretty colour."""
@@ -170,24 +164,18 @@ class Interpreter(code.InteractiveInterpreter):
         exc_type, value, sys.last_traceback = sys.exc_info()
         sys.last_type = exc_type
         sys.last_value = value
-        if filename and exc_type is SyntaxError:
-            # Work hard to stuff the correct filename in the exception
-            try:
-                msg, (dummy_filename, lineno, offset, line) = value.args
-            except:
-                # Not the format we expect; leave it alone
-                pass
-            else:
-                # Stuff in the right filename and right lineno
-                # strip linecache line number
-                if self.bpython_input_re.match(filename):
-                    filename = "<input>"
-                value = SyntaxError(msg, (filename, lineno, offset, line))
-                sys.last_value = value
+        if filename and exc_type is SyntaxError and value is not None:
+            msg = value.args[0]
+            args = list(value.args[1])
+            # strip linechache line number
+            if self.bpython_input_re.match(filename):
+                args[0] = "<input>"
+            value = SyntaxError(msg, tuple(args))
+            sys.last_value = value
         exc_formatted = traceback.format_exception_only(exc_type, value)
         self.writetb(exc_formatted)
 
-    def showtraceback(self):
+    def showtraceback(self) -> None:
         """This needs to override the default traceback thing
         so it can put it into a pretty colour and maybe other
         stuff, I don't know"""
@@ -199,11 +187,10 @@ class Interpreter(code.InteractiveInterpreter):
             tblist = traceback.extract_tb(tb)
             del tblist[:1]
 
-            for i, (fname, lineno, module, something) in enumerate(tblist):
-                # strip linecache line number
-                if self.bpython_input_re.match(fname):
-                    fname = "<input>"
-                    tblist[i] = (fname, lineno, module, something)
+            for frame in tblist:
+                if self.bpython_input_re.match(frame.filename):
+                    # strip linecache line number
+                    frame.filename = "<input>"
 
             l = traceback.format_list(tblist)
             if l:
@@ -214,7 +201,7 @@ class Interpreter(code.InteractiveInterpreter):
 
         self.writetb(l)
 
-    def writetb(self, lines):
+    def writetb(self, lines: Iterable[str]) -> None:
         """This outputs the traceback and should be overridden for anything
         fancy."""
         for line in lines:
@@ -230,78 +217,80 @@ class MatchesIterator:
     A MatchesIterator can be `clear`ed to reset match iteration, and
     `update`ed to set what matches will be iterated over."""
 
-    def __init__(self):
+    def __init__(self) -> None:
         # word being replaced in the original line of text
         self.current_word = ""
         # possible replacements for current_word
-        self.matches = None
+        self.matches: List[str] = []
         # which word is currently replacing the current word
         self.index = -1
         # cursor position in the original line
-        self.orig_cursor_offset = None
+        self.orig_cursor_offset = -1
         # original line (before match replacements)
-        self.orig_line = None
+        self.orig_line = ""
         # class describing the current type of completion
-        self.completer = None
+        self.completer: Optional[autocomplete.BaseCompletionType] = None
+        self.start: Optional[int] = None
+        self.end: Optional[int] = None
 
-    def __nonzero__(self):
+    def __nonzero__(self) -> bool:
         """MatchesIterator is False when word hasn't been replaced yet"""
         return self.index != -1
 
-    def __bool__(self):
+    def __bool__(self) -> bool:
         return self.index != -1
 
     @property
-    def candidate_selected(self):
+    def candidate_selected(self) -> bool:
         """True when word selected/replaced, False when word hasn't been
         replaced yet"""
         return bool(self)
 
-    def __iter__(self):
+    def __iter__(self) -> "MatchesIterator":
         return self
 
-    def current(self):
+    def current(self) -> str:
         if self.index == -1:
             raise ValueError("No current match.")
         return self.matches[self.index]
 
-    def next(self):
-        return self.__next__()
-
-    def __next__(self):
+    def __next__(self) -> str:
         self.index = (self.index + 1) % len(self.matches)
         return self.matches[self.index]
 
-    def previous(self):
+    def previous(self) -> str:
         if self.index <= 0:
             self.index = len(self.matches)
         self.index -= 1
 
         return self.matches[self.index]
 
-    def cur_line(self):
+    def cur_line(self) -> Tuple[int, str]:
         """Returns a cursor offset and line with the current substitution
         made"""
         return self.substitute(self.current())
 
-    def substitute(self, match):
+    def substitute(self, match: str) -> Tuple[int, str]:
         """Returns a cursor offset and line with match substituted in"""
-        start, end, word = self.completer.locate(
-            self.orig_cursor_offset, self.orig_line
-        )
+        assert self.completer is not None
+
+        lp = self.completer.locate(self.orig_cursor_offset, self.orig_line)
+        assert lp is not None
         return (
-            start + len(match),
-            self.orig_line[:start] + match + self.orig_line[end:],
+            lp.start + len(match),
+            self.orig_line[: lp.start] + match + self.orig_line[lp.stop :],
         )
 
-    def is_cseq(self):
+    def is_cseq(self) -> bool:
         return bool(
             os.path.commonprefix(self.matches)[len(self.current_word) :]
         )
 
-    def substitute_cseq(self):
+    def substitute_cseq(self) -> Tuple[int, str]:
         """Returns a new line by substituting a common sequence in, and update
         matches"""
+        assert self.completer is not None
+
         cseq = os.path.commonprefix(self.matches)
         new_cursor_offset, new_line = self.substitute(cseq)
         if len(self.matches) == 1:
@@ -314,7 +303,13 @@ class MatchesIterator:
                 self.clear()
         return new_cursor_offset, new_line
 
-    def update(self, cursor_offset, current_line, matches, completer):
+    def update(
+        self,
+        cursor_offset: int,
+        current_line: str,
+        matches: List[str],
+        completer: autocomplete.BaseCompletionType,
+    ) -> None:
         """Called to reset the match index and update the word being replaced
 
         Should only be called if there's a target to update - otherwise, call
@@ -328,42 +323,73 @@ class MatchesIterator:
         self.matches = matches
         self.completer = completer
         self.index = -1
-        self.start, self.end, self.current_word = self.completer.locate(
-            self.orig_cursor_offset, self.orig_line
-        )
+        lp = self.completer.locate(self.orig_cursor_offset, self.orig_line)
+        assert lp is not None
+        self.start = lp.start
+        self.end = lp.stop
+        self.current_word = lp.word
 
-    def clear(self):
+    def clear(self) -> None:
         self.matches = []
-        self.cursor_offset = -1
-        self.current_line = ""
+        self.orig_cursor_offset = -1
+        self.orig_line = ""
         self.current_word = ""
         self.start = None
         self.end = None
         self.index = -1
 
 
-class Interaction:
-    def __init__(self, config, statusbar=None):
+class Interaction(metaclass=abc.ABCMeta):
+    def __init__(self, config: Config):
         self.config = config
 
-        if statusbar:
-            self.statusbar = statusbar
+    @abc.abstractmethod
+    def confirm(self, s: str) -> bool:
+        pass
 
-    def confirm(self, s):
-        raise NotImplementedError
+    @abc.abstractmethod
+    def notify(
+        self, s: str, n: float = 10.0, wait_for_keypress: bool = False
+    ) -> None:
+        pass
 
-    def notify(self, s, n=10, wait_for_keypress=False):
-        raise NotImplementedError
+    @abc.abstractmethod
+    def file_prompt(self, s: str) -> Optional[str]:
+        pass
 
-    def file_prompt(self, s):
-        raise NotImplementedError
+
+class NoInteraction(Interaction):
+    def __init__(self, config: Config):
+        super().__init__(config)
+
+    def confirm(self, s: str) -> bool:
+        return False
+
+    def notify(
+        self, s: str, n: float = 10.0, wait_for_keypress: bool = False
+    ) -> None:
+        pass
+
+    def file_prompt(self, s: str) -> Optional[str]:
+        return None
 
 
 class SourceNotFound(Exception):
     """Exception raised when the requested source could not be found."""
 
 
-class Repl:
+@dataclass
+class _FuncExpr:
+    """Stack element in Repl._funcname_and_argnum"""
+
+    full_expr: str
+    function_expr: str
+    arg_number: int
+    opening: str
+    keyword: Optional[str] = None
+
+
+class Repl(metaclass=abc.ABCMeta):
     """Implements the necessary guff for a Python-repl-alike interface
 
     The execution of the code entered and all that stuff was taken from the
@@ -396,17 +422,64 @@ class Repl:
     XXX Subclasses should implement echo, current_line, cw
     """
 
-    def __init__(self, interp, config):
+    @abc.abstractmethod
+    def reevaluate(self):
+        pass
+
+    @abc.abstractmethod
+    def reprint_line(
+        self, lineno: int, tokens: List[Tuple[_TokenType, str]]
+    ) -> None:
+        pass
+
+    @abc.abstractmethod
+    def _get_current_line(self) -> str:
+        pass
+
+    @abc.abstractmethod
+    def _set_current_line(self, val: str) -> None:
+        pass
+
+    @property
+    def current_line(self) -> str:
+        """The current line"""
+        return self._get_current_line()
+
+    @current_line.setter
+    def current_line(self, value: str) -> None:
+        self._set_current_line(value)
+
+    @abc.abstractmethod
+    def _get_cursor_offset(self) -> int:
+        pass
+
+    @abc.abstractmethod
+    def _set_cursor_offset(self, val: int) -> None:
+        pass
+
+    @property
+    def cursor_offset(self) -> int:
+        """The current cursor offset from the front of the "line"."""
+        return self._get_cursor_offset()
+
+    @cursor_offset.setter
+    def cursor_offset(self, value: int) -> None:
+        self._set_cursor_offset(value)
+
+    if TYPE_CHECKING:
+        # not actually defined, subclasses must define
+        cpos: int
+
+    def __init__(self, interp: Interpreter, config: Config):
         """Initialise the repl.
 
         interp is a Python code.InteractiveInterpreter instance
 
         config is a populated bpython.config.Struct.
         """
-
         self.config = config
         self.cut_buffer = ""
-        self.buffer = []
+        self.buffer: List[str] = []
         self.interp = interp
         self.interp.syntaxerror_callback = self.clear_current_line
         self.match = False
@@ -415,18 +488,21 @@ class Repl:
         )
         # all input and output, stored as old style format strings
         # (\x01, \x02, ...) for cli.py
-        self.screen_hist = []
-        self.history = []  # commands executed since beginning of session
-        self.redo_stack = []
+        self.screen_hist: List[str] = []
+        # commands executed since beginning of session
+        self.history: List[str] = []
+        self.redo_stack: List[str] = []
         self.evaluating = False
         self.matches_iter = MatchesIterator()
         self.funcprops = None
-        self.arg_pos = None
+        self.arg_pos: Union[str, int, None] = None
         self.current_func = None
-        self.highlighted_paren = None
-        self._C = {}
-        self.prev_block_finished = 0
-        self.interact = Interaction(self.config)
+        self.highlighted_paren: Optional[
+            Tuple[Any, List[Tuple[_TokenType, str]]]
+        ] = None
+        self._C: Dict[str, int] = {}
+        self.prev_block_finished: int = 0
+        self.interact: Interaction = NoInteraction(self.config)
         # previous pastebin content to prevent duplicate pastes, filled on call
         # to repl.pastebin
         self.prev_pastebin_content = ""
@@ -435,11 +511,13 @@ class Repl:
         # Necessary to fix mercurial.ui.ui expecting sys.stderr to have this
         # attribute
         self.closed = False
+        self.paster: Union[PasteHelper, PastePinnwand]
 
         if self.config.hist_file.exists():
             try:
                 self.rl_history.load(
-                    self.config.hist_file, getpreferredencoding() or "ascii"
+                    self.config.hist_file,
+                    getpreferredencoding() or "ascii",
                 )
             except OSError:
                 pass
@@ -466,7 +544,7 @@ class Repl:
     def ps2(self) -> str:
         return cast(str, getattr(sys, "ps2", "... "))
 
-    def startup(self):
+    def startup(self) -> None:
         """
         Execute PYTHONSTARTUP file if it exits. Call this after front
         end-specific initialisation.
@@ -476,7 +554,7 @@ class Repl:
             encoding = inspection.get_encoding_file(filename)
             with open(filename, encoding=encoding) as f:
                 source = f.read()
-                self.interp.runsource(source, filename, "exec", encode=False)
+                self.interp.runsource(source, filename, "exec")
 
     def current_string(self, concatenate=False):
         """If the line ends in a string get it, otherwise return ''"""
@@ -490,7 +568,7 @@ class Repl:
             return ""
         opening = string_tokens.pop()[1]
         string = list()
-        for (token, value) in reversed(string_tokens):
+        for token, value in reversed(string_tokens):
             if token is Token.Text:
                 continue
             elif opening is None:
@@ -509,46 +587,45 @@ class Repl:
             return ""
         return "".join(string)
 
-    def get_object(self, name):
+    def get_object(self, name: str) -> Any:
         attributes = name.split(".")
-        obj = eval(attributes.pop(0), self.interp.locals)
+        obj = eval(attributes.pop(0), cast(Dict[str, Any], self.interp.locals))
         while attributes:
-            with inspection.AttrCleaner(obj):
-                obj = getattr(obj, attributes.pop(0))
+            obj = inspection.getattr_safe(obj, attributes.pop(0))
         return obj
 
     @classmethod
-    def _funcname_and_argnum(cls, line):
+    def _funcname_and_argnum(
+        cls, line: str
+    ) -> Tuple[Optional[str], Optional[Union[str, int]]]:
         """Parse out the current function name and arg from a line of code."""
-        # each list in stack:
-        # [full_expr, function_expr, arg_number, opening]
-        # arg_number may be a string if we've encountered a keyword
-        # argument so we're done counting
-        stack = [["", "", 0, ""]]
+        # each element in stack is a _FuncExpr instance
+        # if keyword is not None, we've encountered a keyword and so we're done counting
+        stack = [_FuncExpr("", "", 0, "")]
         try:
-            for (token, value) in Python3Lexer().get_tokens(line):
+            for token, value in Python3Lexer().get_tokens(line):
                 if token is Token.Punctuation:
                     if value in "([{":
-                        stack.append(["", "", 0, value])
+                        stack.append(_FuncExpr("", "", 0, value))
                     elif value in ")]}":
-                        full, _, _, start = stack.pop()
-                        expr = start + full + value
-                        stack[-1][1] += expr
-                        stack[-1][0] += expr
+                        element = stack.pop()
+                        expr = element.opening + element.full_expr + value
+                        stack[-1].function_expr += expr
+                        stack[-1].full_expr += expr
                     elif value == ",":
-                        try:
-                            stack[-1][2] += 1
-                        except TypeError:
-                            stack[-1][2] = ""
-                        stack[-1][1] = ""
-                        stack[-1][0] += value
-                    elif value == ":" and stack[-1][3] == "lambda":
-                        expr = stack.pop()[0] + ":"
-                        stack[-1][1] += expr
-                        stack[-1][0] += expr
+                        if stack[-1].keyword is None:
+                            stack[-1].arg_number += 1
+                        else:
+                            stack[-1].keyword = ""
+                        stack[-1].function_expr = ""
+                        stack[-1].full_expr += value
+                    elif value == ":" and stack[-1].opening == "lambda":
+                        expr = stack.pop().full_expr + ":"
+                        stack[-1].function_expr += expr
+                        stack[-1].full_expr += expr
                     else:
-                        stack[-1][1] = ""
-                        stack[-1][0] += value
+                        stack[-1].function_expr = ""
+                        stack[-1].full_expr += value
                 elif (
                     token is Token.Number
                     or token in Token.Number.subtypes
@@ -557,25 +634,25 @@ class Repl:
                     or token is Token.Operator
                     and value == "."
                 ):
-                    stack[-1][1] += value
-                    stack[-1][0] += value
+                    stack[-1].function_expr += value
+                    stack[-1].full_expr += value
                 elif token is Token.Operator and value == "=":
-                    stack[-1][2] = stack[-1][1]
-                    stack[-1][1] = ""
-                    stack[-1][0] += value
+                    stack[-1].keyword = stack[-1].function_expr
+                    stack[-1].function_expr = ""
+                    stack[-1].full_expr += value
                 elif token is Token.Number or token in Token.Number.subtypes:
-                    stack[-1][1] = value
-                    stack[-1][0] += value
+                    stack[-1].function_expr = value
+                    stack[-1].full_expr += value
                 elif token is Token.Keyword and value == "lambda":
-                    stack.append([value, "", 0, value])
+                    stack.append(_FuncExpr(value, "", 0, value))
                 else:
-                    stack[-1][1] = ""
-                    stack[-1][0] += value
-            while stack[-1][3] in "[{":
+                    stack[-1].function_expr = ""
+                    stack[-1].full_expr += value
+            while stack[-1].opening in "[{":
                 stack.pop()
-            _, _, arg_number, _ = stack.pop()
-            _, func, _, _ = stack.pop()
-            return func, arg_number
+            elem1 = stack.pop()
+            elem2 = stack.pop()
+            return elem2.function_expr, elem1.keyword or elem1.arg_number
         except IndexError:
             return None, None
 
@@ -616,7 +693,6 @@ class Repl:
                     # py3
                     f.__new__.__class__ is not object.__new__.__class__
                 ):
-
                     class_f = f.__new__
 
                 if class_f:
@@ -636,12 +712,12 @@ class Repl:
         self.arg_pos = None
         return False
 
-    def get_source_of_current_name(self):
+    def get_source_of_current_name(self) -> str:
         """Return the unicode source code of the object which is bound to the
         current name in the current input line. Throw `SourceNotFound` if the
         source cannot be found."""
 
-        obj = self.current_func
+        obj: Optional[Callable] = self.current_func
         try:
             if obj is None:
                 line = self.current_line
@@ -649,7 +725,8 @@ class Repl:
                     raise SourceNotFound(_("Nothing to get source of"))
                 if inspection.is_eval_safe_name(line):
                     obj = self.get_object(line)
-            return inspection.get_source_unicode(obj)
+            # Ignoring the next mypy error because we want this to fail if obj is None
+            return inspect.getsource(obj)  # type:ignore[arg-type]
         except (AttributeError, NameError) as e:
             msg = _("Cannot get source: %s") % (e,)
         except OSError as e:
@@ -661,7 +738,7 @@ class Repl:
                 msg = _("No source code found for %s") % (self.current_line,)
         raise SourceNotFound(msg)
 
-    def set_docstring(self):
+    def set_docstring(self) -> None:
         self.docstring = None
         if not self.get_args():
             self.funcprops = None
@@ -686,7 +763,7 @@ class Repl:
     # If exactly one match that is equal to current line, clear matches
     # If example one match and tab=True, then choose that and clear matches
 
-    def complete(self, tab=False):
+    def complete(self, tab: bool = False) -> Optional[bool]:
         """Construct a full list of possible completions and
         display them in a window. Also check if there's an available argspec
         (via the inspect module) and bang that on top of the completions too.
@@ -705,7 +782,7 @@ class Repl:
             self.completers,
             cursor_offset=self.cursor_offset,
             line=self.current_line,
-            locals_=self.interp.locals,
+            locals_=cast(Dict[str, Any], self.interp.locals),
             argspec=self.funcprops,
             current_block="\n".join(self.buffer + [self.current_line]),
             complete_magic_methods=self.config.complete_magic_methods,
@@ -716,28 +793,33 @@ class Repl:
             self.matches_iter.clear()
             return bool(self.funcprops)
 
-        self.matches_iter.update(
-            self.cursor_offset, self.current_line, matches, completer
-        )
+        if completer:
+            self.matches_iter.update(
+                self.cursor_offset, self.current_line, matches, completer
+            )
 
-        if len(matches) == 1:
-            if tab:
-                # if this complete is being run for a tab key press, substitute
-                # common sequence
-                (
-                    self._cursor_offset,
-                    self._current_line,
-                ) = self.matches_iter.substitute_cseq()
-                return Repl.complete(self)  # again for
-            elif self.matches_iter.current_word == matches[0]:
-                self.matches_iter.clear()
-                return False
-            return completer.shown_before_tab
+            if len(matches) == 1:
+                if tab:
+                    # if this complete is being run for a tab key press, substitute
+                    # common sequence
+                    (
+                        self._cursor_offset,
+                        self._current_line,
+                    ) = self.matches_iter.substitute_cseq()
+                    return Repl.complete(self)  # again for
+                elif self.matches_iter.current_word == matches[0]:
+                    self.matches_iter.clear()
+                    return False
+                return completer.shown_before_tab
 
+            else:
+                return tab or completer.shown_before_tab
         else:
-            return tab or completer.shown_before_tab
+            return False
 
-    def format_docstring(self, docstring, width, height):
+    def format_docstring(
+        self, docstring: str, width: int, height: int
+    ) -> List[str]:
         """Take a string and try to format it into a sane list of strings to be
         put into the suggestion box."""
 
@@ -757,7 +839,7 @@ class Repl:
         out[-1] = out[-1].rstrip()
         return out
 
-    def next_indentation(self):
+    def next_indentation(self) -> int:
         """Return the indentation of the next line based on the current
         input buffer."""
         if self.buffer:
@@ -798,7 +880,7 @@ class Repl:
 
         return "\n".join(process())
 
-    def write2file(self):
+    def write2file(self) -> None:
         """Prompt for a filename and write the current contents of the stdout
         buffer to disk."""
 
@@ -811,25 +893,22 @@ class Repl:
             self.interact.notify(_("Save cancelled."))
             return
 
-        fn = Path(fn).expanduser()
-        if fn.suffix != ".py" and self.config.save_append_py:
+        path = Path(fn).expanduser()
+        if path.suffix != ".py" and self.config.save_append_py:
             # fn.with_suffix(".py") does not append if fn has a non-empty suffix
-            fn = Path(f"{fn}.py")
+            path = Path(f"{path}.py")
 
         mode = "w"
-        if fn.exists():
-            mode = self.interact.file_prompt(
+        if path.exists():
+            new_mode = self.interact.file_prompt(
                 _(
-                    "%s already exists. Do you "
-                    "want to (c)ancel, "
-                    " (o)verwrite or "
-                    "(a)ppend? "
+                    "%s already exists. Do you want to (c)ancel, (o)verwrite or (a)ppend? "
                 )
-                % (fn,)
+                % (path,)
             )
-            if mode in ("o", "overwrite", _("overwrite")):
+            if new_mode in ("o", "overwrite", _("overwrite")):
                 mode = "w"
-            elif mode in ("a", "append", _("append")):
+            elif new_mode in ("a", "append", _("append")):
                 mode = "a"
             else:
                 self.interact.notify(_("Save cancelled."))
@@ -838,14 +917,14 @@ class Repl:
         stdout_text = self.get_session_formatted_for_file()
 
         try:
-            with open(fn, mode) as f:
+            with open(path, mode) as f:
                 f.write(stdout_text)
         except OSError as e:
-            self.interact.notify(_("Error writing file '%s': %s") % (fn, e))
+            self.interact.notify(_("Error writing file '%s': %s") % (path, e))
         else:
-            self.interact.notify(_("Saved to %s.") % (fn,))
+            self.interact.notify(_("Saved to %s.") % (path,))
 
-    def copy2clipboard(self):
+    def copy2clipboard(self) -> None:
         """Copy current content to clipboard."""
 
         if not have_pyperclip:
@@ -860,7 +939,7 @@ class Repl:
         else:
             self.interact.notify(_("Copied content to clipboard."))
 
-    def pastebin(self, s=None):
+    def pastebin(self, s=None) -> Optional[str]:
         """Upload to a pastebin and display the URL in the status bar."""
 
         if s is None:
@@ -870,11 +949,13 @@ class Repl:
             _("Pastebin buffer? (y/N) ")
         ):
             self.interact.notify(_("Pastebin aborted."))
+            return None
         else:
             return self.do_pastebin(s)
 
-    def do_pastebin(self, s):
+    def do_pastebin(self, s) -> Optional[str]:
         """Actually perform the upload."""
+        paste_url: str
         if s == self.prev_pastebin_content:
             self.interact.notify(
                 _("Duplicate pastebin. Previous URL: %s. " "Removal URL: %s")
@@ -888,11 +969,11 @@ class Repl:
             paste_url, removal_url = self.paster.paste(s)
         except PasteFailed as e:
             self.interact.notify(_("Upload failed: %s") % e)
-            return
+            return None
 
         self.prev_pastebin_content = s
         self.prev_pastebin_url = paste_url
-        self.prev_removal_url = removal_url
+        self.prev_removal_url = removal_url if removal_url is not None else ""
 
         if removal_url is not None:
             self.interact.notify(
@@ -905,7 +986,7 @@ class Repl:
 
         return paste_url
 
-    def push(self, s, insert_into_history=True):
+    def push(self, s, insert_into_history=True) -> bool:
         """Push a line of code onto the buffer so it can process it all
         at once when a code block ends"""
         # This push method is used by cli and urwid, but not curtsies
@@ -915,14 +996,14 @@ class Repl:
         if insert_into_history:
             self.insert_into_history(s)
 
-        more = self.interp.runsource("\n".join(self.buffer))
+        more: bool = self.interp.runsource("\n".join(self.buffer))
 
         if not more:
             self.buffer = []
 
         return more
 
-    def insert_into_history(self, s):
+    def insert_into_history(self, s: str):
         try:
             self.rl_history.append_reload_and_write(
                 s, self.config.hist_file, getpreferredencoding()
@@ -930,7 +1011,7 @@ class Repl:
         except RuntimeError as e:
             self.interact.notify(f"{e}")
 
-    def prompt_undo(self):
+    def prompt_undo(self) -> int:
         """Returns how many lines to undo, 0 means don't undo"""
         if (
             self.config.single_undo_time < 0
@@ -938,14 +1019,18 @@ class Repl:
         ):
             return 1
         est = self.interp.timer.estimate()
-        n = self.interact.file_prompt(
+        m = self.interact.file_prompt(
             _("Undo how many lines? (Undo will take up to ~%.1f seconds) [1]")
             % (est,)
         )
+        if m is None:
+            self.interact.notify(_("Undo canceled"), 0.1)
+            return 0
+
         try:
-            if n == "":
-                n = "1"
-            n = int(n)
+            if m == "":
+                m = "1"
+            n = int(m)
         except ValueError:
             self.interact.notify(_("Undo canceled"), 0.1)
             return 0
@@ -962,7 +1047,7 @@ class Repl:
                 self.interact.notify(message % (n, est), 0.1)
             return n
 
-    def undo(self, n=1):
+    def undo(self, n: int = 1) -> None:
         """Go back in the undo history n steps and call reevaluate()
         Note that in the program this is called "Rewind" because I
         want it to be clear that this is by no means a true undo
@@ -986,7 +1071,7 @@ class Repl:
 
         self.rl_history.entries = entries
 
-    def flush(self):
+    def flush(self) -> None:
         """Olivier Grisel brought it to my attention that the logging
         module tries to call this method, since it makes assumptions
         about stdout that may not necessarily be true. The docs for
@@ -1003,7 +1088,7 @@ class Repl:
     def close(self):
         """See the flush() method docstring."""
 
-    def tokenize(self, s, newline=False):
+    def tokenize(self, s, newline=False) -> List[Tuple[_TokenType, str]]:
         """Tokenizes a line of code, returning pygments tokens
         with side effects/impurities:
         - reads self.cpos to see what parens should be highlighted
@@ -1020,7 +1105,7 @@ class Repl:
         cursor = len(source) - self.cpos
         if self.cpos:
             cursor += 1
-        stack = list()
+        stack: List[Any] = list()
         all_tokens = list(Python3Lexer().get_tokens(source))
         # Unfortunately, Pygments adds a trailing newline and strings with
         # no size, so strip them
@@ -1029,10 +1114,10 @@ class Repl:
         all_tokens[-1] = (all_tokens[-1][0], all_tokens[-1][1].rstrip("\n"))
         line = pos = 0
         parens = dict(zip("{([", "})]"))
-        line_tokens = list()
-        saved_tokens = list()
+        line_tokens: List[Tuple[_TokenType, str]] = list()
+        saved_tokens: List[Tuple[_TokenType, str]] = list()
         search_for_paren = True
-        for (token, value) in split_lines(all_tokens):
+        for token, value in split_lines(all_tokens):
             pos += len(value)
             if token is Token.Text and value == "\n":
                 line += 1
@@ -1101,11 +1186,11 @@ class Repl:
             return list()
         return line_tokens
 
-    def clear_current_line(self):
+    def clear_current_line(self) -> None:
         """This is used as the exception callback for the Interpreter instance.
         It prevents autoindentation from occurring after a traceback."""
 
-    def send_to_external_editor(self, text):
+    def send_to_external_editor(self, text: str) -> str:
         """Returns modified text from an editor, or the original text if editor
         exited with non-zero"""
 
@@ -1163,34 +1248,22 @@ class Repl:
             self.interact.notify(_("Error editing config file: %s") % e)
 
 
-def next_indentation(line, tab_length):
+def next_indentation(line, tab_length) -> int:
     """Given a code line, return the indentation of the next line."""
     line = line.expandtabs(tab_length)
-    indentation = (len(line) - len(line.lstrip(" "))) // tab_length
+    indentation: int = (len(line) - len(line.lstrip(" "))) // tab_length
     if line.rstrip().endswith(":"):
         indentation += 1
     elif indentation >= 1:
-        if line.lstrip().startswith(("return", "pass", "raise", "yield")):
+        if line.lstrip().startswith(
+            ("return", "pass", "...", "raise", "yield", "break", "continue")
+        ):
             indentation -= 1
     return indentation
 
 
-def next_token_inside_string(code_string, inside_string):
-    """Given a code string s and an initial state inside_string, return
-    whether the next token will be inside a string or not."""
-    for token, value in Python3Lexer().get_tokens(code_string):
-        if token is Token.String:
-            value = value.lstrip("bBrRuU")
-            if value in ('"""', "'''", '"', "'"):
-                if not inside_string:
-                    inside_string = value
-                elif value == inside_string:
-                    inside_string = False
-    return inside_string
-
-
 def split_lines(tokens):
-    for (token, value) in tokens:
+    for token, value in tokens:
         if not value:
             continue
         while value:
